import pika
from pika.exchange_type import ExchangeType
from propzen.common.service_layer.externalbus import IExternalPublisher, ExternalEvent
from propzen.common.adapters.rabbitmq.connection import Connection


class RabbitExternalPublisher(IExternalPublisher):

    def __init__(self, conn: Connection, servicename: str):
        """
        Parameters
        ----------
        conn : Connection
            RabbitMQ Connection from propzen-common
        servicename: str
            Service name of publishing origin
        """

        self.connection = conn
        self.servicename = servicename

    def publish(self, event: ExternalEvent):
        # TODO: implement retry if failed to publish
        with self.connection.connect() as channel:
            channel.exchange_declare(
                exchange=Connection.exchange,
                exchange_type=ExchangeType.direct,
                durable=True)

            channel.basic_publish(
                exchange=Connection.exchange,
                routing_key=event.routingkey(),
                body=event.to_json(),
                properties=pika.BasicProperties(
                    delivery_mode=pika.spec.PERSISTENT_DELIVERY_MODE))



from uuid import UUID
from dataclasses import dataclass
from propzen.common.service_layer.externalbus import ExternalEvent


@dataclass
class AssetAttachmentRemoved(ExternalEvent):
    account_id: UUID
    asset_id: UUID
    file_id: UUID


@dataclass
class AssetImageRemoved(ExternalEvent):
    account_id: UUID
    asset_id: UUID
    file_id: UUID



import os
from typing import Any
from pika.exchange_type import ExchangeType
from pika.exceptions import ConnectionClosedByBroker, AMQPChannelError, AMQPConnectionError
from tenacity import retry_if_exception_type, wait_fixed, Retrying, RetryCallState

from propzen.common.service_layer.externalbus import ExternalEventHandler
from propzen.common.adapters.rabbitmq.connection import Connection


class RabbitConsumer:
    exchange: str = 'propzen.direct'

    def __init__(self, queue_name: str, bus: Connection, logger: Any):
        self.queue_name = queue_name
        self.bus = bus
        self.handlers = {}
        self.types = {}
        self.logger = logger
        self.is_running = False

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.dispose()

    def run_consume_with_retry(self):
        def reconnect_before(retry_state: RetryCallState):
            self.logger.info('Reconnecting...')
            self.dispose()

        retry_consume = Retrying(
            retry=(
                retry_if_exception_type(ConnectionClosedByBroker) |
                retry_if_exception_type(AMQPChannelError) |
                retry_if_exception_type(AMQPConnectionError)),
            wait=wait_fixed(5),
            before=reconnect_before,
        )

        retry_consume(self.run_consume)

    def run_consume(self):
        try:
            self.logger.info('Starting consumer...')

            self.bus.open_channel()
            self.logger.info('Channel opened...')

            self.bus.channel.exchange_declare(
                exchange=RabbitConsumer.exchange,
                exchange_type=ExchangeType.direct,
                durable=True)
            self.logger.info('Exchange declared...')

            self.bus.channel.queue_declare(
                queue=self.queue_name,
                durable=True)
            self.logger.info(f'Queue {self.queue_name} declared...')

            self._bindhandlers()

            self.bus.channel.basic_consume(
                queue=self.queue_name,
                on_message_callback=self._process_event)

            self.bus.channel.start_consuming()
            self.is_running = True
            self.logger.info('Consumer started...')
        except KeyboardInterrupt:
            self.dispose()
            self.logger.info('Closing rabbitmq consumer...')
            os._exit(0)

    def dispose(self):
        if self.is_running:
            self._unbindhandlers()
            self.bus.channel.stop_consuming()
        self.bus.close()
        self.is_running = False

    def add_handlers(self, handlers: ExternalEventHandler):
        for eventtype, handler in handlers.mapping.items():
            eventname = eventtype.routingkey()
            self.types[eventname] = eventtype
            self.handlers[eventname] = handler

    def _process_event(self, channel, method, properties, body):
        eventname = method.routing_key
        eventjson = body.decode('utf-8')
        event = self.types[eventname].from_json(eventjson)
        self.logger.info(
            'external event received',
            externalevent=method.routing_key,
            **event.to_dict())
        try:
            self.handlers[eventname](event)
            channel.basic_ack(delivery_tag=method.delivery_tag)
        except Exception as exc:
            self.logger.exception(exc, externalevent=method.routing_key)

    def _bindhandlers(self):
        for eventname in self.handlers:
            self.bus.channel.queue_bind(
                exchange=RabbitConsumer.exchange,
                queue=self.queue_name,
                routing_key=eventname
            )
            self.logger.info(
                f'{eventname} event bound to {self.queue_name}.')

    def _unbindhandlers(self):
        for eventname in self.handlers:
            self.bus.channel.queue_unbind(
                queue=self.queue_name,
                exchange=RabbitConsumer.exchange,
                routing_key=eventname)



from __future__ import annotations
from os import environ
from typing import Union, Optional, Any, Type, TypeVar
from pathlib import Path
from logging import getLogger
from pydantic import BaseSettings, Field
from pydantic.env_settings import read_env_file


class Settings(BaseSettings):
    def __init__(self, base_dir: Union[Path, str], _env_file: Union[Path, str, None] = ..., _env_file_encoding: Optional[str] = None,
                 _secrets_dir: Union[Path, str, None] = None, **values: Any) -> None:
        super().__init__(_env_file=_env_file, _env_file_encoding=_env_file_encoding,
                         _secrets_dir=_secrets_dir, **values)
        self._base_dir = base_dir

    # EXECUTION ENVIRONMENT
    exec_env: str = Field('development', env='EXEC_ENV')

    @property
    def configuration(self):
        return self.exec_env.lower()

    # PATH
    @property
    def base_dir(self) -> str:
        return self._base_dir

    # TEST
    test_run: bool = False
    test_env_value: str = ''  # for unit test config

    # APP DETAILS
    project_name: str = 'PropZen'

    # DATABASE
    db_url: str = 'postgres://user:pass@localhost:5432/propzen'

    # MESSAGE BUS
    bus_url: str = 'amqp://localhost:5672'

    def override_with_dict(self, overrides: dict):
        def env_and_original():
            for key in self.__fields__.keys():
                env = self.__fields__[key].field_info.extra.get('env')
                env_or_key = env or key
                if not self.__config__.case_sensitive:
                    env_or_key = env_or_key.lower()
                yield env_or_key, key
        [
            setattr(self, original, overrides[env])
            for (env, original) in env_and_original()
            if env in overrides.keys() and overrides[env] is not None
        ]

    def override_with_envfile(self, env_path: Path):
        if env_path.exists():
            overrides = read_env_file(env_path)
            self.override_with_dict(overrides)
            getLogger().info('Loaded %s environment overrides.', env_path)

    class Config:
        case_sensitive = False
        extra = 'allow'


SettingsType = TypeVar('SettingsType', bound=Settings)


def init_settings(
    base_dir: Union[Path, str],
    settings_cls: Type[SettingsType] = Settings
) -> SettingsType:

    def get_configuration_path(env_filename: str):
        return Path(base_dir, 'configurations', env_filename or '')

    default = settings_cls(
        base_dir=base_dir,
        _env_file=get_configuration_path('.env'),
        _env_file_encoding='utf-8')

    configuration = default.configuration
    env_filename = f'.env.{configuration}'

    try:
        default.override_with_envfile(
            env_path=get_configuration_path(env_filename))

        env_filename = f'.env.{configuration}.local'
        default.override_with_envfile(
            env_path=get_configuration_path(env_filename))

        env_filename = '.env.local'
        default.override_with_envfile(
            env_path=get_configuration_path(env_filename))

        environ_dict = dict(environ)
        if default.__config__.case_sensitive:
            default.override_with_dict(environ_dict)
        else:
            default.override_with_dict(
                {k.lower(): v for k, v in environ_dict.items()})

        default = settings_cls(
            base_dir=base_dir,
            _env_file=get_configuration_path('.env'),
            _env_file_encoding='utf-8',
            **default.dict())

    except Exception as exc:
        getLogger().exception(
            'Exception thrown when trying to load %s environment overrides.', env_filename,
            exc_info=exc
        )

    return default


class ApiSettings(Settings):
    jwt_secret_key: str
    jwt_algorithm: str
    jwt_access_expire_minutes: int
    jwt_refresh_expire_minutes: int
    oauth2_token_url: str



from uuid import uuid4, UUID

from .event import Event


class Entity():

    def __init__(self):
        self.id = uuid4()
        self.events = []

    def __eq__(self, other) -> bool:
        if isinstance(other, self.__class__):
            return self.id == other.id
        return False

    def __hash__(self) -> int:
        return hash(self.id)
